@@ -2,12 +2,12 @@ import h5py
 import hdfdict
 import jax.numpy as jnp
 import jax.ops
-import numpy as np
 import scipy as scipy
 import scipy.integrate
 import scipy.optimize
 from jax import jit, partial
 from jax.config import config
+import numpy as np
 import pkg_resources
 
 USE_JAX = True
@@ -37,7 +37,7 @@ class Box:
         getter method for the name of the Box Class.
 
         :return:
-            returns a string that represents the name of the Box.
+        returns a string that represents the name of the Box.
         """
         return self._name
 
@@ -46,7 +46,7 @@ class Box:
         getter method for the reservoir content of the Box Class.
 
         :return:
-            returns the reservoir content that was passed into __init__ method.
+        returns the reservoir content that was passed into __init__ method.
         """
         return self._reservoir
 
@@ -55,16 +55,16 @@ class Box:
         getter method for the reservoir content of the Box Class.
 
         :return:
-            returns the production coefficient that was passed into __init__ method.
+        returns the production coefficient that was passed into __init__ method.
         """
         return self._production
 
     def __str__(self):
         """
-       returns a string representation of the Box.
+        returns a string representation of the Box.
 
         :return:
-            returns the string in thw follow representation - name:reservoir_content:production_value
+        returns the string in thw follow representation - name:reservoir_content:production_value
         """
 
         return self._name + ":" + str(self._reservoir) + ":" + str(self._production)
@@ -73,40 +73,51 @@ class Box:
 class Flow:
     def __init__(self, source, destination, flux_rate):
         """
+        init method for when the Flow object is created.
 
-        :param source:
-        :param destination:
-        :param flux_rate:
+        :param source: source of the flow.
+        :param destination: destination the flow.
+        :param flux_rate: flux rate between the source and the destination. Flux must be a non-negative value.
         """
         self._source = source
         self._destination = destination
+        assert flux_rate >= 0
         self._flux = flux_rate
 
     def get_source(self):
         """
+        getter method for the source node of the Flow Class.
 
         :return:
+        returns the source of the flow.
+
         """
         return self._source
 
     def get_destination(self):
         """
+        getter method for the destination node of the Flow Class.
 
         :return:
+        returns the destination of the flow.
         """
         return self._destination
 
     def get_flux(self):
         """
+        getter method for the flux rate of the Flow Class.
 
         :return:
+        returns the flux of the flow.
         """
         return self._flux
 
     def __str__(self):
         """
+        returns a string representation of the Flow.
 
         :return:
+        returns the string in thw follow representation - str(source) --> str(destination):flux_value
         """
         return str(self._source) + " --> " + str(self._destination) + " : " + str(self._flux)
 
@@ -114,9 +125,11 @@ class Flow:
 class CarbonBoxModel:
     def __init__(self, production_rate_units='kg/yr', flow_rate_units='Gt/yr'):
         """
+        init method for when the Carbon Box Model object is created. Initialises the parameters that are needed for
+        the rest of the class.
 
-        :param production_rate_units:
-        :param flow_rate_units:
+        :param production_rate_units: units for the production rate. Only valid values are 'kg/yr' or 'atoms/cm^2/s'.
+        :param flow_rate_units: units for the flow rate. Only valid values are 'Gt/yr' or '1/yr'.
         """
         self._nodes = {}
         self._reverse_nodes = {}
@@ -126,11 +139,12 @@ class CarbonBoxModel:
         self._fluxes = None
         self._decay_matrix = None
         self._production_coefficients = None
-        self._decay_constant = jax.numpy.log(2) / 5730
+        self._decay_constant = jnp.log(2) / 5730
         self._production_rate_units = production_rate_units
         self._flow_rate_units = flow_rate_units
         self._corrected_fluxes = None
         self._matrix = None
+        self._growth_kernel = jnp.array([1] * 12)
 
     def add_nodes(self, nodes):
         """
@@ -294,68 +308,22 @@ class CarbonBoxModel:
         else:
             raise ValueError("Must give either target C-14 or production rate.")
 
-    # def production_function_generator(self, data, time):
-    #     mse = np.inf
-    #     troposphere_index = None
-    #     for index, node in self._nodes.items():
-    #         if node.get_name() == 'Troposphere':
-    #             troposphere_index = index
-    #             break
-    #     if troposphere_index is None:
-    #         raise ValueError('there is currently no Troposphere node to equilibrate!')
-    #
-    #     equilibrium = self.equilibrate(production_rate=self.equilibrate(target_C_14=data[0]))
-    #     production_terms = self._reverse_pass(data, time, troposphere_index, equilibrium=equilibrium)
-    #
-    #     while mse > 0.001:
-    #         forwards = self._forward_pass(time, equilibrium, production_terms)
-    #         mse = np.sum((forwards[troposphere_index, :] - data) ** 2)
-    #         production_terms = self._reverse_pass(data, time, troposphere_index, first_pass=False,
-    #                                               box_contents=forwards)
-    #         print(mse)
-    #
-    #     return production_terms
-    #
-    # def _forward_pass(self, time, equilibrium, production_rates):
-    #     production_fun = (lambda x: (jnp.interp(x, time, production_rates)))
-    #     binned_data = self.run_bin(time, 1000, production_fun, y0=equilibrium)
-    #     return binned_data
-    #
-    # def _reverse_pass(self, troposphere_content, time, troposphere_index, first_pass=True, box_contents=None,
-    #                   equilibrium=None):
-    #     production_terms = []
-    #     time_step = time[1] - time[0]
-    #     if first_pass:
-    #         for i in range(len(time) - 1):
-    #             new_y = np.concatenate([equilibrium[:troposphere_index], np.array([troposphere_content[i]]),
-    #                                     equilibrium[troposphere_index + 1:]])
-    #             production_term = (troposphere_content[i + 1] - troposphere_content[i]) / time_step \
-    #                               - self._matrix[troposphere_index, :] * new_y
-    #             production_terms.append(production_term / 0.3)
-    #     else:
-    #         for i in range(len(time) - 1):
-    #             new_box_contents = box_contents.copy()
-    #             new_box_contents[troposphere_index, :] = troposphere_content
-    #             # new_y = box_contents[i,:troposphere_index] + troposphere_content[i] + \
-    #             #         box_contents[i,troposphere_index+1:]
-    #             # print(new_y)
-    #             production_term = (troposphere_content[i + 1] - troposphere_content[i]) / time_step \
-    #                               - self._matrix[troposphere_index, :] * new_box_contents[:, i]
-    #             production_terms.append(production_term / 0.3)
-    #     production_terms.append(0)
-    #     return production_terms
-
     def run(self, time_values, production, y0=None, args=(), target_C_14=None, steady_state_production=None):
 
         @jit
         def derivative(y, t):
             ans = jnp.matmul(self._matrix, y)
+            # print("t shape: ", t.shape)
+            # print("y: ", y)
+            # print("t: ", t)
+            # print("args: ", *args)
             production_rate_constant = production(t, *args)
             production_rate_constant = self._convert_production_rate(production_rate_constant)
             production_term = self._production_coefficients * production_rate_constant
             return ans + production_term
 
         time_values = jnp.array(time_values)
+        # print("time_values shape: ", time_values.shape)
         solution = None
         if y0 is not None:
             y_initial = jnp.array(y0)
@@ -366,42 +334,57 @@ class CarbonBoxModel:
             elif target_C_14 is not None:
                 solution = self.equilibrate(production_rate=self.equilibrate(target_C_14=target_C_14))
             else:
-                ValueError("Must give either target C-14 or production rate.")
+                ValueError("Must give either y0, or target C-14, or production rate.")
             y_initial = jnp.array(solution)
 
         if not callable(production):
             raise ValueError("incorrect object type for production")
 
         if USE_JAX:
+            # print("time_values jax shape: ", time_values.shape)
             states = odeint(derivative, y_initial, time_values)
         else:
+            # print("time_values no jax shape: ", time_values.shape)
             states = odeint(derivative, y_initial, time_values)
         return states, solution
 
     def run_bin(self, time_out, time_oversample, production, y0=None, args=(), target_C_14=None,
                 steady_state_production=None):
+        # time_out = jnp.array(time_out)
+        # # time_step = time_out[1] - time_out[0]
+        # t = jnp.linspace(jnp.min(time_out), jnp.max(time_out), (time_out.shape[0] - 1) * time_oversample)
+        # states, solution = self.run(t, production, y0=y0, args=args, target_C_14=target_C_14,
+        #                             steady_state_production=steady_state_production)
+        # m = int(1 / time_step * time_oversample // 12)
+        # tiled = jnp.resize(jnp.repeat(self._growth_kernel, m), (1,  int(time_oversample / time_step)))
+        # num = int(time_oversample // (time_step * 12) * 12)
+        # tiled = jax.ops.index_update(tiled, jnp.array([[False] * num + [True] * (tiled.shape[1] - num)]), 0)
+        # tiled_full = jnp.resize(jnp.tile(tiled, (states.shape[1], int((time_out.shape[0] - 1) * time_step))),
+        #                         (states.shape[1], states.shape[0]))
+        # # print(tiled_full.shape)
+        # # print(states)
+        # states = jnp.transpose(tiled_full) * states
+        # # print(jnp.reshape(states, (-1, states.shape[0] // time_oversample, time_oversample, states.shape[1])).sum(2))
+        # binned_data = jnp.reshape(states, (-1, states.shape[0] // time_oversample, time_oversample, states.shape[1])) \
+        #                   .sum(2).sum(0) / jnp.sum(tiled) / time_step
+
         time_out = jnp.array(time_out)
         t = jnp.linspace(jnp.min(time_out), jnp.max(time_out), (time_out.shape[0] - 1) * time_oversample)
         states, solution = self.run(t, production, y0=y0, args=args, target_C_14=target_C_14,
                                     steady_state_production=steady_state_production)
+        m = time_oversample // 12
+        tiled = jnp.resize(jnp.repeat(self._growth_kernel, m), (1,  time_oversample))
+        num = int((time_oversample // 12) * 12)
+        tiled = jax.ops.index_update(tiled, jnp.array([[False] * num + [True] * (tiled.shape[1] - num)]), 0)
+        tiled_full = jnp.resize(jnp.tile(tiled, (states.shape[1], int(time_out.shape[0] - 1))),
+                                (states.shape[1], states.shape[0]))
+        states = jnp.transpose(tiled_full) * states
         binned_data = jnp.reshape(states, (-1, states.shape[0] // time_oversample, time_oversample, states.shape[1])) \
-                          .sum(2).sum(0) / time_oversample
+                          .sum(2).sum(0) / jnp.sum(tiled)
 
         return binned_data, solution
 
-    def run_D_14_C_values(self, time_out, time_oversample, production, y0=None, args=(), target_C_14=None,
-                          steady_state_production=None, steady_state_solutions=None):
-
-        time_out = jnp.array(time_out)
-        data, soln = self.run_bin(time_out=time_out, time_oversample=time_oversample, production=production,
-                                  y0=y0, args=args, target_C_14=target_C_14,
-                                  steady_state_production=steady_state_production)
-
-        if steady_state_solutions is None:
-            solution = soln
-
-        else:
-            solution = steady_state_solutions
+    def _to_d14c(self,data,solution):
 
         troposphere_steady_state = None
         d_14_c = None
@@ -415,6 +398,35 @@ class CarbonBoxModel:
         return d_14_c
 
 
+    def run_D_14_C_values(self, time_out, time_oversample, production, y0=None, args=(), target_C_14=None,
+                          steady_state_production=None, steady_state_solutions=None):
+        time_out = jnp.array(time_out)
+        data, soln = self.run_bin(time_out=time_out, time_oversample=time_oversample, production=production,
+                                  y0=y0, args=args, target_C_14=target_C_14,
+                                  steady_state_production=steady_state_production)
+
+        if steady_state_solutions is None:
+            solution = soln
+        else:
+            solution = steady_state_solutions
+
+        return self._to_d14c(data,solution)
+
+    def define_growth_season(self, months):
+        month_list = np.array(['january', 'february', 'march', 'april', 'may', 'june', 'july', 'august', 'september',
+                               'october', 'november', 'december'])
+        months = np.array(months)
+        self._growth_kernel = jax.ops.index_update(self._growth_kernel, np.in1d(month_list, months,
+                                                                                invert=True), 0)
+
+    def define_growth_season(self, months):
+        month_list = np.array(['january', 'february', 'march', 'april', 'may', 'june', 'july', 'august', 'september',
+                               'october', 'november', 'december'])
+        months = np.array(months)
+        self._growth_kernel = jax.ops.index_update(self._growth_kernel, np.in1d(month_list, months,
+                                                                                invert=True), 0)
+
+
 def save_model(carbon_box_model, filename):
     file = h5py.File(filename, 'a')
     file.clear()  # overwrites if it already exists
